@@ -39,6 +39,9 @@
 # define MIME_UI_DEFAULT 0
 #endif
 
+/* Registry path to store plugin settings */
+#define GPGOL_REGPATH "Software\\GNU\\GpgOL"
+
 #ifdef __cplusplus
 extern "C" {
 #if 0
@@ -60,6 +63,11 @@ const char *default_homedir (void);
 char *get_data_dir (void);
 char *get_gpg4win_dir (void);
 
+int store_extension_value (const char *key, const char *val);
+int store_extension_subkey_value (const char *subkey, const char *key,
+                                  const char *val);
+int load_extension_value (const char *key, char **val);
+
 /* Get a temporary filename with and its name */
 wchar_t *get_tmp_outfile (wchar_t *name, HANDLE *outHandle);
 
@@ -81,8 +89,6 @@ const char *get_pubkey_algo_str (gpgme_pubkey_algo_t id);
 
 /*-- config-dialog.c --*/
 void config_dialog_box (HWND parent);
-int store_extension_value (const char *key, const char *val);
-int load_extension_value (const char *key, char **val);
 
 /*-- verify-dialog.c --*/
 int verify_dialog_box (gpgme_protocol_t protocol,
@@ -140,6 +146,19 @@ extern int g_ol_version_major;
 void log_window_hierarchy (HWND window, const char *fmt,
                            ...) __attribute__ ((format (printf,2,3)));
 
+void bring_to_front (HWND wid);
+
+int gpgol_message_box (HWND parent, const char *utf8_text,
+                       const char *utf8_caption, UINT type);
+
+/* Show a bug message with the code. */
+void gpgol_bug (HWND parent, int code);
+
+#define ERR_CRYPT_RESOLVER_FAILED 1
+#define ERR_WANTS_SEND_MIME_BODY 2
+#define ERR_WANTS_SEND_INLINE_BODY 3
+#define ERR_INLINE_BODY_TO_BODY 4
+#define ERR_INLINE_BODY_INV_STATE 5
 #ifdef __cplusplus
 }
 #endif
