@@ -33,9 +33,13 @@
 #include "gpgoladdin.h"
 #include "mymapitags.h"
 #include "parsecontroller.h"
+#include "cryptcontroller.h"
 #include "gpgolstr.h"
 #include "windowmessages.h"
 #include "mlang-charset.h"
+#include "wks-helper.h"
+#include "keycache.h"
+#include "cpphelp.h"
 
 #include <gpgme++/configuration.h>
 #include <gpgme++/tofuinfo.h>
@@ -57,61 +61,11 @@
 
 using namespace GpgME;
 
-static std::map<LPDISPATCH, Mail*> g_mail_map;
-static std::map<std::string, Mail*> g_uid_map;
+static std::map<LPDISPATCH, Mail*> s_mail_map;
+static std::map<std::string, Mail*> s_uid_map;
 static std::set<std::string> uids_searched;
 
-static bool
-in_de_vs_mode()
-{
-  /* We cache the values only once. A change requires restart.
-     This is because checking this is very expensive as gpgconf
-     spawns each process to query the settings. */
-  static bool checked;
-  static bool vs_mode;
-
-  if (checked)
-    {
-      return vs_mode;
-    }
-  Error err;
-  const auto components = Configuration::Component::load (err);
-  log_debug ("%s:%s: Checking for de-vs mode.",
-             SRCNAME, __func__);
-  if (err)
-    {
-      log_error ("%s:%s: Failed to get gpgconf components: %s",
-                 SRCNAME, __func__, err.asString ());
-      checked = true;
-      vs_mode = false;
-      return vs_mode;
-    }
-  for (const auto &component: components)
-    {
-      if (component.name () && !strcmp (component.name (), "gpg"))
-        {
-          for (const auto &option: component.options ())
-            {
-              if (option.name () && !strcmp (option.name (), "compliance") &&
-                  option.currentValue ().stringValue () &&
-                  !stricmp (option.currentValue ().stringValue (), "de-vs"))
-                {
-                  log_debug ("%s:%s: Detected de-vs mode",
-                             SRCNAME, __func__);
-                  checked = true;
-                  vs_mode = true;
-                  return vs_mode;
-                }
-            }
-          checked = true;
-          vs_mode = false;
-          return vs_mode;
-        }
-    }
-  checked = true;
-  vs_mode = false;
-  return false;
-}
+static Mail *s_last_mail;
 
 #define COPYBUFSIZE (8 * 1024)
 
@@ -135,7 +89,13 @@ Mail::Mail (LPDISPATCH mailitem) :
     m_cached_recipients(nullptr),
     m_type(MSGTYPE_UNKNOWN),
     m_do_inline(false),
-    m_is_gsuite(false)
+    m_is_gsuite(false),
+    m_crypt_state(NoCryptMail),
+    m_window(nullptr),
+    m_async_crypt_disabled(false),
+    m_is_forwarded_crypto_mail(false),
+    m_is_send_again(false),
+    m_disable_att_remove_warning(false)
 {
   if (get_mail_for_item (mailitem))
     {
@@ -154,7 +114,8 @@ Mail::Mail (LPDISPATCH mailitem) :
       gpgol_release(mailitem);
       return;
     }
-  g_mail_map.insert (std::pair<LPDISPATCH, Mail *> (mailitem, this));
+  s_mail_map.insert (std::pair<LPDISPATCH, Mail *> (mailitem, this));
+  s_last_mail = this;
 }
 
 GPGRT_LOCK_DEFINE(dtor_lock);
@@ -167,27 +128,38 @@ Mail::~Mail()
      that the parser is alive even if the mail is deleted
      while parsing. */
   gpgrt_lock_lock (&dtor_lock);
+  log_oom_extra ("%s:%s: dtor: Mail: %p item: %p",
+                 SRCNAME, __func__, this, m_mailitem);
   std::map<LPDISPATCH, Mail *>::iterator it;
 
+  log_oom_extra ("%s:%s: Detaching event sink",
+                 SRCNAME, __func__);
   detach_MailItemEvents_sink (m_event_sink);
   gpgol_release(m_event_sink);
 
-  it = g_mail_map.find(m_mailitem);
-  if (it != g_mail_map.end())
+  log_oom_extra ("%s:%s: Erasing mail",
+                 SRCNAME, __func__);
+  it = s_mail_map.find(m_mailitem);
+  if (it != s_mail_map.end())
     {
-      g_mail_map.erase (it);
+      s_mail_map.erase (it);
     }
 
   if (!m_uuid.empty())
     {
-      auto it2 = g_uid_map.find(m_uuid);
-      if (it2 != g_uid_map.end())
+      auto it2 = s_uid_map.find(m_uuid);
+      if (it2 != s_uid_map.end())
         {
-          g_uid_map.erase (it2);
+          s_uid_map.erase (it2);
         }
     }
 
+  log_oom_extra ("%s:%s: releasing mailitem",
+                 SRCNAME, __func__);
   gpgol_release(m_mailitem);
+  xfree (m_cached_html_body);
+  xfree (m_cached_plain_body);
+  release_cArray (m_cached_recipients);
   if (!m_uuid.empty())
     {
       log_oom_extra ("%s:%s: destroyed: %p uuid: %s",
@@ -195,15 +167,16 @@ Mail::~Mail()
     }
   else
     {
-      log_oom_extra ("%s:%s: non crypto mail: %p destroyed",
+      log_oom_extra ("%s:%s: non crypto (or sent) mail: %p destroyed",
                      SRCNAME, __func__, this);
     }
-  xfree (m_cached_html_body);
-  xfree (m_cached_plain_body);
-  for (int i = 0; m_cached_recipients && m_cached_recipients[i]; ++i)
-      xfree (m_cached_recipients[i]);
-  xfree (m_cached_recipients);
+  log_oom_extra ("%s:%s: nulling shared pointer",
+                 SRCNAME, __func__);
+  m_parser = nullptr;
+  m_crypter = nullptr;
   gpgrt_lock_unlock (&dtor_lock);
+  log_oom_extra ("%s:%s: returning",
+                 SRCNAME, __func__);
 }
 
 Mail *
@@ -214,8 +187,8 @@ Mail::get_mail_for_item (LPDISPATCH mailitem)
       return NULL;
     }
   std::map<LPDISPATCH, Mail *>::iterator it;
-  it = g_mail_map.find(mailitem);
-  if (it == g_mail_map.end())
+  it = s_mail_map.find(mailitem);
+  if (it == s_mail_map.end())
     {
       return NULL;
     }
@@ -229,8 +202,8 @@ Mail::get_mail_for_uuid (const char *uuid)
     {
       return NULL;
     }
-  auto it = g_uid_map.find(std::string(uuid));
-  if (it == g_uid_map.end())
+  auto it = s_uid_map.find(std::string(uuid));
+  if (it == s_uid_map.end())
     {
       return NULL;
     }
@@ -240,8 +213,8 @@ Mail::get_mail_for_uuid (const char *uuid)
 bool
 Mail::is_valid_ptr (const Mail *mail)
 {
-  auto it = g_mail_map.begin();
-  while (it != g_mail_map.end())
+  auto it = s_mail_map.begin();
+  while (it != s_mail_map.end())
     {
       if (it->second == mail)
         return true;
@@ -660,6 +633,7 @@ add_attachments(LPDISPATCH mail,
                      SRCNAME, __func__, att->get_display_name().c_str());
           err = 1;
         }
+      CloseHandle (hFile);
       if (!DeleteFileW (wchar_file))
         {
           log_error ("%s:%s: Failed to delete tmp attachment for: %s",
@@ -696,6 +670,7 @@ do_parsing (LPVOID arg)
   gpgrt_lock_unlock (&dtor_lock);
 
   gpgrt_lock_lock (&parser_lock);
+  gpgrt_lock_lock (&invalidate_lock);
   /* We lock the parser here to avoid too many
      decryption attempts if there are
      multiple mailobjects which might have already
@@ -709,6 +684,7 @@ do_parsing (LPVOID arg)
     {
       log_debug ("%s:%s: cancel for: %p already deleted",
                  SRCNAME, __func__, arg);
+      gpgrt_lock_unlock (&invalidate_lock);
       gpgrt_lock_unlock (&parser_lock);
       return 0;
     }
@@ -717,15 +693,197 @@ do_parsing (LPVOID arg)
     {
       log_error ("%s:%s: no parser found for mail: %p",
                  SRCNAME, __func__, arg);
+      gpgrt_lock_unlock (&invalidate_lock);
       gpgrt_lock_unlock (&parser_lock);
       return -1;
     }
   parser->parse();
   do_in_ui_thread (PARSING_DONE, arg);
+  gpgrt_lock_unlock (&invalidate_lock);
   gpgrt_lock_unlock (&parser_lock);
   return 0;
 }
 
+/* How encryption is done:
+
+   There are two modes of encryption. Synchronous and Async.
+   If async is used depends on the value of mail->async_crypt_disabled.
+
+   Synchronous crypto:
+
+   > Send Event < | State NoCryptMail
+   Needs Crypto ? (get_gpgol_draft_info_flags != 0)
+
+   -> No:
+      Pass send -> unencrypted mail.
+
+   -> Yes:
+      mail->update_oom_data
+      State = Mail::NeedsFirstAfterWrite
+      check_inline_response
+      invoke_oom_method (m_object, "Save", NULL);
+
+      > Write Event <
+      Pass because is_crypto_mail is false (not a decrypted mail)
+
+      > AfterWrite Event < | State NeedsFirstAfterWrite
+      State = NeedsActualCrypo
+      encrypt_sign_start
+        collect_input_data
+        -> Check if Inline PGP should be used
+        do_crypt
+          -> Resolve keys / do crypto
+
+          State = NeedsUpdateInMAPI
+          update_crypt_mapi
+          crypter->update_mail_mapi
+           if (inline) (Meaning PGP/Inline)
+          <-- do nothing.
+           else
+            build MSOXSMIME attachment and clear body / attachments.
+
+          State = NeedsUpdateInOOM
+      <- Back to Send Event
+      update_crypt_oom
+        -> Cleans body or sets PGP/Inline body. (inline_body_to_body)
+      State = WantsSendMIME or WantsSendInline
+
+      -> Saftey check "has_crypted_or_empty_body"
+      -> If MIME Mail do the T3656 check.
+
+    Send.
+
+    State order for "inline_response" (sync) Mails.
+    NoCryptMail
+    NeedsFirstAfterWrite
+    NeedsActualCrypto
+    NeedsUpdateInMAPI
+    NeedsUpdateInOOM
+    WantsSendMIME (or inline for PGP Inline)
+    -> Send.
+
+    State order for async Mails
+    NoCryptMail
+    NeedsFirstAfterWrite
+    NeedsActualCrypto
+    -> Cancel Send.
+    Windowmessages -> Crypto Done
+    NeedsUpdateInOOM
+    NeedsSecondAfterWrite
+    trigger Save.
+    NeedsUpdateInMAPI
+    WantsSendMIME
+    trigger Send.
+*/
+static DWORD WINAPI
+do_crypt (LPVOID arg)
+{
+  gpgrt_lock_lock (&dtor_lock);
+  /* We lock with mail dtors so we can be sure the mail->parser
+     call is valid. */
+  Mail *mail = (Mail *)arg;
+  if (!Mail::is_valid_ptr (mail))
+    {
+      log_debug ("%s:%s: canceling crypt for: %p already deleted",
+                 SRCNAME, __func__, arg);
+      gpgrt_lock_unlock (&dtor_lock);
+      return 0;
+    }
+  if (mail->crypt_state() != Mail::NeedsActualCrypt)
+    {
+      log_debug ("%s:%s: invalid state %i",
+                 SRCNAME, __func__, mail->crypt_state ());
+      mail->set_window_enabled (true);
+      gpgrt_lock_unlock (&dtor_lock);
+      return -1;
+    }
+
+  /* This takes a shared ptr of crypter. So the crypter is
+     still valid when the mail is deleted. */
+  auto crypter = mail->crypter();
+  gpgrt_lock_unlock (&dtor_lock);
+
+  if (!crypter)
+    {
+      log_error ("%s:%s: no crypter found for mail: %p",
+                 SRCNAME, __func__, arg);
+      gpgrt_lock_unlock (&parser_lock);
+      mail->set_window_enabled (true);
+      return -1;
+    }
+
+  int rc = crypter->do_crypto();
+
+  gpgrt_lock_lock (&dtor_lock);
+  if (!Mail::is_valid_ptr (mail))
+    {
+      log_debug ("%s:%s: aborting crypt for: %p already deleted",
+                 SRCNAME, __func__, arg);
+      gpgrt_lock_unlock (&dtor_lock);
+      return 0;
+    }
+
+  if (rc == -1)
+    {
+      mail->reset_crypter ();
+      crypter = nullptr;
+      gpgol_bug (mail->get_window (),
+                 ERR_CRYPT_RESOLVER_FAILED);
+    }
+
+  mail->set_window_enabled (true);
+
+  if (rc)
+    {
+      log_debug ("%s:%s: crypto failed for: %p with: %i",
+                 SRCNAME, __func__, arg, rc);
+      mail->set_crypt_state (Mail::NoCryptMail);
+      mail->reset_crypter ();
+      crypter = nullptr;
+      gpgrt_lock_unlock (&dtor_lock);
+      return rc;
+    }
+
+  if (!mail->async_crypt_disabled ())
+    {
+      mail->set_crypt_state (Mail::NeedsUpdateInOOM);
+      gpgrt_lock_unlock (&dtor_lock);
+      // This deletes the Mail in Outlook 2010
+      do_in_ui_thread (CRYPTO_DONE, arg);
+      log_debug ("%s:%s: UI thread finished for %p",
+                 SRCNAME, __func__, arg);
+    }
+  else
+    {
+      mail->set_crypt_state (Mail::NeedsUpdateInMAPI);
+      mail->update_crypt_mapi ();
+      if (mail->crypt_state () == Mail::WantsSendMIME)
+        {
+          // For sync crypto we need to switch this.
+          mail->set_crypt_state (Mail::NeedsUpdateInOOM);
+        }
+      else
+        {
+          // A bug!
+          log_debug ("%s:%s: Resetting crypter because of state mismatch. %p",
+                     SRCNAME, __func__, arg);
+          crypter = nullptr;
+          mail->reset_crypter ();
+        }
+      gpgrt_lock_unlock (&dtor_lock);
+    }
+  /* This works around a bug in pinentry that it might
+     bring the wrong window to front. So after encryption /
+     signing we bring outlook back to front.
+
+     See GnuPG-Bug-Id: T3732
+     */
+  do_in_ui_thread_async (BRING_TO_FRONT, nullptr);
+  log_debug ("%s:%s: crypto thread for %p finished",
+             SRCNAME, __func__, arg);
+  return 0;
+}
+
 bool
 Mail::is_crypto_mail() const
 {
@@ -755,9 +913,22 @@ Mail::decrypt_verify()
     }
   set_uuid ();
   m_processed = true;
+
+
   /* Insert placeholder */
   char *placeholder_buf;
-  if (gpgrt_asprintf (&placeholder_buf, opt.prefer_html ? decrypt_template_html :
+  if (m_type == MSGTYPE_GPGOL_WKS_CONFIRMATION)
+    {
+      gpgrt_asprintf (&placeholder_buf, opt.prefer_html ? decrypt_template_html :
+                      decrypt_template,
+                      "OpenPGP",
+                      _("Pubkey directory confirmation"),
+                      _("This is a confirmation request to publish your Pubkey in the "
+                        "directory for your domain.\n\n"
+                        "<p>If you did not request to publish your Pubkey in your providers "
+                        "directory, simply ignore this message.</p>\n"));
+    }
+  else if (gpgrt_asprintf (&placeholder_buf, opt.prefer_html ? decrypt_template_html :
                       decrypt_template,
                       is_smime() ? "S/MIME" : "OpenPGP",
                       _("Encrypted message"),
@@ -791,6 +962,12 @@ Mail::decrypt_verify()
   /* Do the actual parsing */
   auto cipherstream = get_attachment_stream (m_mailitem, m_moss_position);
 
+  if (m_type == MSGTYPE_GPGOL_WKS_CONFIRMATION)
+    {
+      WKSHelper::instance ()->handle_confirmation_read (this, cipherstream);
+      return 0;
+    }
+
   if (!cipherstream)
     {
       log_debug ("%s:%s: Failed to get cipherstream.",
@@ -830,6 +1007,12 @@ void find_and_replace(std::string& source, const std::string &find,
 void
 Mail::update_body()
 {
+  if (!m_parser)
+    {
+      TRACEPOINT;
+      return;
+    }
+
   const auto error = m_parser->get_formatted_error ();
   if (!error.empty())
     {
@@ -841,6 +1024,12 @@ Mail::update_body()
               log_error ("%s:%s: Failed to modify html body of item.",
                          SRCNAME, __func__);
             }
+          else
+            {
+              log_debug ("%s:%s: Set error html to: '%s'",
+                         SRCNAME, __func__, error.c_str ());
+            }
+
         }
       else
         {
@@ -850,6 +1039,11 @@ Mail::update_body()
               log_error ("%s:%s: Failed to modify html body of item.",
                          SRCNAME, __func__);
             }
+          else
+            {
+              log_debug ("%s:%s: Set error plain to: '%s'",
+                         SRCNAME, __func__, error.c_str ());
+            }
         }
       return;
     }
@@ -952,7 +1146,7 @@ Mail::parsing_done()
     }
 
   update_sigstate ();
-  m_needs_wipe = true;
+  m_needs_wipe = !m_is_send_again;
 
   TRACEPOINT;
   /* Set categories according to the result. */
@@ -973,19 +1167,44 @@ Mail::parsing_done()
                  SRCNAME, __func__);
     }
 
-  /* Invalidate UI to set the correct sig status. */
-  m_parser = nullptr;
-  gpgoladdin_invalidate_ui ();
+  if (m_is_send_again)
+    {
+      log_debug ("%s:%s: I think that this is the send again of a crypto mail.",
+                 SRCNAME, __func__);
+
+      /* We no longer want to be treated like a crypto mail. */
+      m_type = MSGTYPE_UNKNOWN;
+      LPMESSAGE msg = get_oom_base_message (m_mailitem);
+      if (!msg)
+        {
+          TRACEPOINT;
+        }
+      else
+        {
+          set_gpgol_draft_info_flags (msg, m_crypto_flags);
+          gpgol_release (msg);
+        }
+      remove_our_attachments ();
+    }
+
+  log_debug ("%s:%s: Delayed invalidate to update sigstate.",
+             SRCNAME, __func__);
+  CloseHandle(CreateThread (NULL, 0, delayed_invalidate_ui, (LPVOID) this, 0,
+                            NULL));
   TRACEPOINT;
   return;
 }
 
 int
-Mail::encrypt_sign ()
+Mail::encrypt_sign_start ()
 {
-  int err = -1,
-      flags = 0;
-  protocol_t proto = opt.enable_smime ? PROTOCOL_UNKNOWN : PROTOCOL_OPENPGP;
+  if (m_crypt_state != NeedsActualCrypt)
+    {
+      log_debug ("%s:%s: invalid state %i",
+                 SRCNAME, __func__, m_crypt_state);
+      return -1;
+    }
+  int flags = 0;
   if (!needs_crypto())
     {
       return 0;
@@ -995,9 +1214,10 @@ Mail::encrypt_sign ()
     {
       log_error ("%s:%s: Failed to get base message.",
                  SRCNAME, __func__);
-      return err;
+      return -1;
     }
   flags = get_gpgol_draft_info_flags (message);
+  gpgol_release (message);
 
   const auto window = get_active_hwnd ();
 
@@ -1008,10 +1228,10 @@ Mail::encrypt_sign ()
       mapi_release_attach_table (att_table);
       /* Check for attachments if we have some abort. */
 
-      wchar_t *w_title = utf8_to_wchar (_(
-                              "GpgOL: Oops, G Suite Sync account detected"));
       if (n_att_usable)
         {
+          wchar_t *w_title = utf8_to_wchar (_(
+                                              "GpgOL: Oops, G Suite Sync account detected"));
           wchar_t *msg = utf8_to_wchar (
                       _("G Suite Sync breaks outgoing crypto mails "
                         "with attachments.\nUsing crypto and attachments "
@@ -1025,78 +1245,37 @@ Mail::encrypt_sign ()
           xfree (w_title);
           return -1;
         }
-      if (flags == 2)
-        {
-          wchar_t *msg = utf8_to_wchar (
-                       _("G Suite Sync breaks outgoing signed mails.\n"
-                        "Ensuring mail integrity (signing) with G Suite Sync "
-                        "is not supported.\n\n"
-                        "See: https://dev.gnupg.org/T3545 for details."));
-          MessageBoxW (window,
-                       msg,
-                       w_title,
-                       MB_ICONINFORMATION|MB_OK);
-          xfree (msg);
-          xfree (w_title);
-          return -1;
-        }
-      if (flags == 3)
-        {
-          wchar_t *msg = utf8_to_wchar (
-                        _("G Suite Sync breaks outgoing signed mails.\n"
-                          "Ensuring mail integrity (signing) with G Suite Sync "
-                          "is not supported.\n\n"
-                          "See: https://dev.gnupg.org/T3545 for details.\n\n"
-                          "Do you want to only encrypt the message?"));
-          if(MessageBoxW (window,
-                          msg,
-                          w_title,
-                          MB_ICONINFORMATION|MB_YESNO) != IDYES)
-            {
-              xfree (msg);
-              xfree (w_title);
-              return -1;
-            }
-          else
-            {
-              flags = 1;
-            }
-          xfree (msg);
-        }
-      xfree (w_title);
     }
 
   m_do_inline = m_is_gsuite ? true : opt.inline_pgp;
 
-  EnableWindow (window, FALSE);
-  if (flags == 3)
+  GpgME::Protocol proto = opt.enable_smime ? GpgME::UnknownProtocol: GpgME::OpenPGP;
+  m_crypter = std::shared_ptr <CryptController> (new CryptController (this, flags & 1,
+                                                                      flags & 2,
+                                                                      proto));
+
+  // Careful from here on we have to check every
+  // error condition with window enabling again.
+  set_window_enabled (false);
+  if (m_crypter->collect_data ())
     {
-      log_debug ("%s:%s: Sign / Encrypting message",
-                 SRCNAME, __func__);
-      err = message_sign_encrypt (message, proto,
-                                  NULL, get_sender ().c_str (), this);
+      log_error ("%s:%s: Crypter for mail %p failed to collect data.",
+                 SRCNAME, __func__, this);
+      set_window_enabled (true);
+      return -1;
     }
-  else if (flags == 2)
+
+  if (!m_async_crypt_disabled)
     {
-      err = message_sign (message, proto,
-                          NULL, get_sender ().c_str (), this);
-    }
-  else if (flags == 1)
-    {
-      err = message_encrypt (message, proto,
-                             NULL, get_sender ().c_str (), this);
+      CloseHandle(CreateThread (NULL, 0, do_crypt,
+                                (LPVOID) this, 0,
+                                NULL));
     }
   else
     {
-      log_debug ("%s:%s: Unknown flags for crypto: %i",
-                 SRCNAME, __func__, flags);
+      do_crypt (this);
     }
-  log_debug ("%s:%s: Status: %i",
-             SRCNAME, __func__, err);
-  EnableWindow (window, TRUE);
-  gpgol_release (message);
-  m_crypt_successful = !err;
-  return err;
+  return 0;
 }
 
 int
@@ -1136,6 +1315,10 @@ Mail::wipe (bool force)
         }
       return -1;
     }
+  else
+    {
+      put_oom_string (m_mailitem, "Body", "");
+    }
   m_needs_wipe = false;
   return 0;
 }
@@ -1143,7 +1326,7 @@ Mail::wipe (bool force)
 int
 Mail::update_oom_data ()
 {
-  LPDISPATCH sender = NULL;
+  char *buf = nullptr;
   log_debug ("%s:%s", SRCNAME, __func__);
 
   if (!is_crypto_mail())
@@ -1162,9 +1345,7 @@ Mail::update_oom_data ()
       xfree (m_cached_plain_body);
       m_cached_plain_body = get_oom_string (m_mailitem, "Body");
 
-      for (int i = 0; m_cached_recipients && m_cached_recipients[i]; ++i)
-          xfree (m_cached_recipients[i]);
-      xfree (m_cached_recipients);
+      release_cArray (m_cached_recipients);
       m_cached_recipients = get_recipients ();
     }
   /* For some reason outlook may store the recipient address
@@ -1177,78 +1358,39 @@ Mail::update_oom_data ()
          you send from the folder of userA but change the from to userB
          outlook will keep the SenderEmailAddress of UserA. This is all
          so horrible. */
-      char *type = get_oom_string (m_mailitem, "SenderEmailType");
-      if (type && !strcmp ("SMTP", type))
-        {
-          char *senderMail = get_oom_string (m_mailitem, "SenderEmailAddress");
-          if (senderMail)
-            {
-              log_debug ("%s:%s Sender found", SRCNAME, __func__);
-              m_sender = senderMail;
-              xfree (senderMail);
-              xfree (type);
-              return 0;
-            }
-        }
-      xfree (type);
-    }
-  sender = get_oom_object (m_mailitem, "SendUsingAccount");
-  if (sender)
-    {
-      char *buf = get_oom_string (sender, "SmtpAddress");
-      char *dispName = get_oom_string (sender, "DisplayName");
-      gpgol_release (sender);
+      buf = get_sender_SenderEMailAddress (m_mailitem);
 
-      /* Check for G Suite account */
-      if (dispName && !strcmp ("G Suite", dispName))
+      if (!buf)
         {
-          m_is_gsuite = true;
+          /* Try the sender Object */
+          buf = get_sender_Sender (m_mailitem);
         }
-      xfree (dispName);
-      if (buf && strlen (buf))
-        {
-          log_debug ("%s:%s Sender fallback 1", SRCNAME, __func__);
-          m_sender = buf;
-          xfree (buf);
-          return 0;
-        }
-      xfree (buf);
-    }
-  /* Fallback to Sender object */
-  sender = get_oom_object (m_mailitem, "Sender");
-  if (sender)
-    {
-      char *buf = get_pa_string (sender, PR_SMTP_ADDRESS_DASL);
-      gpgol_release (sender);
-      if (buf && strlen (buf))
-        {
-          log_debug ("%s:%s Sender fallback 2", SRCNAME, __func__);
-          m_sender = buf;
-          xfree (buf);
-          return 0;
-        }
-      xfree (buf);
-    }
-  /* We don't have s sender object or SendUsingAccount,
-     well, in that case fall back to the current user. */
-  sender = get_oom_object (m_mailitem, "Session.CurrentUser");
-  if (sender)
-    {
-      char *buf = get_pa_string (sender, PR_SMTP_ADDRESS_DASL);
-      gpgol_release (sender);
-      if (buf && strlen (buf))
-        {
-          log_debug ("%s:%s Sender fallback 3", SRCNAME, __func__);
-          m_sender = buf;
-          xfree (buf);
-          return 0;
-        }
-      xfree (buf);
     }
 
-  log_debug ("%s:%s: All fallbacks failed.",
-             SRCNAME, __func__);
-  return -1;
+  if (!buf)
+    {
+      buf = get_sender_SendUsingAccount (m_mailitem, &m_is_gsuite);
+    }
+  if (!buf && !is_crypto_mail ())
+    {
+      /* Try the sender Object */
+      buf = get_sender_Sender (m_mailitem);
+    }
+  if (!buf)
+    {
+      /* We don't have s sender object or SendUsingAccount,
+         well, in that case fall back to the current user. */
+      buf = get_sender_CurrentUser (m_mailitem);
+    }
+  if (!buf)
+    {
+      log_debug ("%s:%s: All fallbacks failed.",
+                 SRCNAME, __func__);
+      return -1;
+    }
+  m_sender = buf;
+  xfree (buf);
+  return 0;
 }
 
 std::string
@@ -1259,13 +1401,19 @@ Mail::get_sender ()
   return m_sender;
 }
 
+std::string
+Mail::get_cached_sender ()
+{
+  return m_sender;
+}
+
 int
 Mail::close_all_mails ()
 {
   int err = 0;
   std::map<LPDISPATCH, Mail *>::iterator it;
   TRACEPOINT;
-  std::map<LPDISPATCH, Mail *> mail_map_copy = g_mail_map;
+  std::map<LPDISPATCH, Mail *> mail_map_copy = s_mail_map;
   for (it = mail_map_copy.begin(); it != mail_map_copy.end(); ++it)
     {
       /* XXX For non racy code the is_valid_ptr check should not
@@ -1299,7 +1447,7 @@ Mail::revert_all_mails ()
 {
   int err = 0;
   std::map<LPDISPATCH, Mail *>::iterator it;
-  for (it = g_mail_map.begin(); it != g_mail_map.end(); ++it)
+  for (it = s_mail_map.begin(); it != s_mail_map.end(); ++it)
     {
       if (it->second->revert ())
         {
@@ -1324,7 +1472,7 @@ Mail::wipe_all_mails ()
 {
   int err = 0;
   std::map<LPDISPATCH, Mail *>::iterator it;
-  for (it = g_mail_map.begin(); it != g_mail_map.end(); ++it)
+  for (it = s_mail_map.begin(); it != s_mail_map.end(); ++it)
     {
       if (it->second->wipe ())
         {
@@ -1344,6 +1492,8 @@ Mail::revert ()
       return 0;
     }
 
+  m_disable_att_remove_warning = true;
+
   err = gpgol_mailitem_revert (m_mailitem);
   if (err == -1)
     {
@@ -1354,6 +1504,7 @@ Mail::revert ()
   /* We need to reprocess the mail next time around. */
   m_processed = false;
   m_needs_wipe = false;
+  m_disable_att_remove_warning = false;
   return 0;
 }
 
@@ -1394,7 +1545,8 @@ Mail::is_smime ()
         }
       else
         {
-          log_error ("Protocol in multipart signed mail.");
+          log_error ("%s:%s: No protocol in multipart / signed mail.",
+                     SRCNAME, __func__);
         }
       xfree (proto);
       xfree (ct);
@@ -1802,7 +1954,7 @@ Mail::set_uuid()
                      SRCNAME, __func__, m_mailitem, uuid);
           delete other;
         }
-      g_uid_map.insert (std::pair<std::string, Mail *> (m_uuid, this));
+      s_uid_map.insert (std::pair<std::string, Mail *> (m_uuid, this));
       log_debug ("%s:%s: uuid for %p is now %s",
                  SRCNAME, __func__, this,
                  m_uuid.c_str());
@@ -2299,71 +2451,15 @@ Mail::get_sig_fpr() const
   return m_sig.fingerprint();
 }
 
-
-static DWORD WINAPI
-do_locate (LPVOID arg)
-{
-  char *recipient = (char*) arg;
-  log_debug ("%s:%s searching key for recipient: \"%s\"",
-             SRCNAME, __func__, recipient);
-  Context *ctx = Context::createForProtocol (OpenPGP);
-
-  if (!ctx)
-    {
-      TRACEPOINT;
-      return 0;
-    }
-
-  ctx->setKeyListMode (GpgME::Extern | GpgME::Local);
-  ctx->startKeyListing (recipient, false);
-
-  std::vector<Key> keys;
-  Error err;
-  do {
-      keys.push_back (ctx->nextKey(err));
-    } while (!err);
-  keys.pop_back ();
-  ctx->endKeyListing ();
-  delete ctx;
-
-  if (keys.size ())
-    {
-      log_debug ("%s:%s found key for recipient: \"%s\"",
-                 SRCNAME, __func__, recipient);
-    }
-  xfree (recipient);
-  // do_in_ui_thread (UNKNOWN, NULL);
-  return 0;
-}
-
-GPGRT_LOCK_DEFINE(uids_searched_lock);
-
 /** Try to locate the keys for all recipients */
-void Mail::locate_keys()
+void
+Mail::locate_keys()
 {
-  gpgrt_lock_lock (&uids_searched_lock);
   char ** recipients = get_recipients ();
-
-  if (!recipients)
-    {
-      TRACEPOINT;
-      return;
-    }
-  for (int i = 0; recipients[i]; i++)
-    {
-      std::string recp = recipients[i];
-      if (uids_searched.find (recp) == uids_searched.end ())
-        {
-          uids_searched.insert (recp);
-          HANDLE thread = CreateThread (NULL, 0, do_locate,
-                                        (LPVOID) strdup(recipients[i]), 0,
-                                        NULL);
-          CloseHandle (thread);
-        }
-      xfree (recipients[i]);
-    }
-  xfree (recipients);
-  gpgrt_lock_unlock (&uids_searched_lock);
+  KeyCache::instance()->startLocate (recipients);
+  KeyCache::instance()->startLocate (get_sender ().c_str ());
+  KeyCache::instance()->startLocateSecret (get_sender ().c_str ());
+  release_cArray (recipients);
 }
 
 bool
@@ -2423,11 +2519,399 @@ Mail::append_to_inline_body (const std::string &data)
 int
 Mail::inline_body_to_body()
 {
-  if (m_inline_body.empty())
+  if (!m_crypter)
+    {
+      log_error ("%s:%s: No crypter.",
+                 SRCNAME, __func__);
+      return -1;
+    }
+
+  const auto body = m_crypter->get_inline_data ();
+  if (body.empty())
     {
       return 0;
     }
-  int ret = put_oom_string (m_mailitem, "Body", m_inline_body.c_str ());
-  m_inline_body = std::string();
+
+  int ret = put_oom_string (m_mailitem, "Body",
+                            body.c_str ());
   return ret;
 }
+
+void
+Mail::update_crypt_mapi()
+{
+  log_debug ("%s:%s: Update crypt mapi",
+             SRCNAME, __func__);
+  if (m_crypt_state != NeedsUpdateInMAPI)
+    {
+      log_debug ("%s:%s: invalid state %i",
+                 SRCNAME, __func__, m_crypt_state);
+      return;
+    }
+  if (!m_crypter)
+    {
+      if (!m_mime_data.empty())
+        {
+          log_debug ("%s:%s: Have override mime data creating dummy crypter",
+                     SRCNAME, __func__);
+          m_crypter = std::shared_ptr <CryptController> (new CryptController (this, false,
+                                                                              false,
+                                                                              GpgME::UnknownProtocol));
+        }
+      else
+        {
+          log_error ("%s:%s: No crypter.",
+                     SRCNAME, __func__);
+          m_crypt_state = NoCryptMail;
+          return;
+        }
+    }
+
+  if (m_crypter->update_mail_mapi ())
+    {
+      log_error ("%s:%s: Failed to update MAPI after crypt",
+                 SRCNAME, __func__);
+      m_crypt_state = NoCryptMail;
+    }
+  else
+    {
+      m_crypt_state = WantsSendMIME;
+    }
+
+  /** If sync we need the crypter in update_crypt_oom */
+  if (!async_crypt_disabled ())
+    {
+      // We don't need the crypter anymore.
+      reset_crypter ();
+    }
+}
+
+/** Checks in OOM if the body is either
+  empty or contains the -----BEGIN tag.
+  pair.first -> true if body starts with -----BEGIN
+  pair.second -> true if body is empty. */
+static std::pair<bool, bool>
+has_crypt_or_empty_body_oom (Mail *mail)
+{
+  auto body = mail->get_body();
+  std::pair<bool, bool> ret;
+  ret.first = false;
+  ret.second = false;
+  ltrim (body);
+  if (body.size() > 10 && !strncmp (body.c_str(), "-----BEGIN", 10))
+    {
+      ret.first = true;
+      return ret;
+    }
+  if (!body.size())
+    {
+      ret.second = true;
+    }
+  else
+    {
+      log_mime_parser ("%s:%s: Body found in %p : \"%s\"",
+                       SRCNAME, __func__, mail, body.c_str ());
+    }
+  return ret;
+}
+
+void
+Mail::update_crypt_oom()
+{
+  log_debug ("%s:%s: Update crypt oom for %p",
+             SRCNAME, __func__, this);
+  if (m_crypt_state != NeedsUpdateInOOM)
+    {
+      log_debug ("%s:%s: invalid state %i",
+                 SRCNAME, __func__, m_crypt_state);
+      reset_crypter ();
+      return;
+    }
+
+  if (do_pgp_inline ())
+    {
+      if (inline_body_to_body ())
+        {
+          log_error ("%s:%s: Inline body to body failed %p.",
+                     SRCNAME, __func__, this);
+          gpgol_bug (get_active_hwnd(), ERR_INLINE_BODY_TO_BODY);
+          m_crypt_state = NoCryptMail;
+          return;
+        }
+    }
+
+  if (m_crypter->get_protocol () == GpgME::CMS && m_crypter->is_encrypter ())
+    {
+      /* We put the PIDNameContentType headers here for exchange
+         because this is the only way we found to inject the
+         smime-type. */
+      if (put_pa_string (m_mailitem,
+                         PR_PIDNameContentType_DASL,
+                         "application/pkcs7-mime;smime-type=\"enveloped-data\";name=smime.p7m"))
+        {
+          log_debug ("%s:%s: Failed to put PIDNameContentType for %p.",
+                     SRCNAME, __func__, this);
+        }
+    }
+
+  /** When doing async update_crypt_mapi follows and needs
+    the crypter. */
+  if (async_crypt_disabled ())
+    {
+      reset_crypter ();
+    }
+
+  const auto pair = has_crypt_or_empty_body_oom (this);
+  if (pair.first)
+    {
+      log_debug ("%s:%s: Looks like inline body. You can pass %p.",
+                 SRCNAME, __func__, this);
+      m_crypt_state = WantsSendInline;
+      return;
+    }
+
+  // We are in MIME land. Wipe the body.
+  if (wipe (true))
+    {
+      log_debug ("%s:%s: Cancel send for %p.",
+                 SRCNAME, __func__, this);
+      wchar_t *title = utf8_to_wchar (_("GpgOL: Encryption not possible!"));
+      wchar_t *msg = utf8_to_wchar (_(
+                                      "Outlook returned an error when trying to send the encrypted mail.\n\n"
+                                      "Please restart Outlook and try again.\n\n"
+                                      "If it still fails consider using an encrypted attachment or\n"
+                                      "switching to PGP/Inline in GpgOL's options."));
+      MessageBoxW (get_active_hwnd(), msg, title,
+                   MB_ICONERROR | MB_OK);
+      xfree (msg);
+      xfree (title);
+      m_crypt_state = NoCryptMail;
+      return;
+    }
+  m_crypt_state = NeedsSecondAfterWrite;
+  return;
+}
+
+void
+Mail::set_window_enabled (bool value)
+{
+  if (!value)
+    {
+      m_window = get_active_hwnd ();
+    }
+  log_debug ("%s:%s: enable window %p %i",
+             SRCNAME, __func__, m_window, value);
+
+  EnableWindow (m_window, value ? TRUE : FALSE);
+}
+
+bool
+Mail::check_inline_response ()
+{
+/* Async sending might lead to crashes when the send invocation is done.
+ * For now we treat every mail as an inline response to disable async
+ * encryption. :-( For more details see: T3838 */
+#ifdef DO_ASYNC_CRYPTO
+  m_async_crypt_disabled = false;
+  LPDISPATCH app = GpgolAddin::get_instance ()->get_application ();
+  if (!app)
+    {
+      TRACEPOINT;
+      return false;
+    }
+
+  LPDISPATCH explorer = get_oom_object (app, "ActiveExplorer");
+
+  if (!explorer)
+    {
+      TRACEPOINT;
+      return false;
+    }
+
+  LPDISPATCH inlineResponse = get_oom_object (explorer, "ActiveInlineResponse");
+  gpgol_release (explorer);
+
+  if (!inlineResponse)
+    {
+      return false;
+    }
+
+  // We have inline response
+  // Check if we are it. It's a bit naive but meh. Worst case
+  // is that we think inline response too often and do sync
+  // crypt where we could do async crypt.
+  char * inlineSubject = get_oom_string (inlineResponse, "Subject");
+  gpgol_release (inlineResponse);
+
+  const auto subject = get_subject ();
+  if (inlineResponse && !subject.empty() && !strcmp (subject.c_str (), inlineSubject))
+    {
+      log_debug ("%s:%s: Detected inline response for '%p'",
+                 SRCNAME, __func__, this);
+      m_async_crypt_disabled = true;
+    }
+  xfree (inlineSubject);
+#else
+  m_async_crypt_disabled = true;
+#endif
+
+  return m_async_crypt_disabled;
+}
+
+// static
+Mail *
+Mail::get_last_mail ()
+{
+  if (!s_last_mail || !is_valid_ptr (s_last_mail))
+    {
+      s_last_mail = nullptr;
+    }
+  return s_last_mail;
+}
+
+// static
+void
+Mail::invalidate_last_mail ()
+{
+  s_last_mail = nullptr;
+}
+
+// static
+void
+Mail::locate_all_crypto_recipients()
+{
+  if (!opt.autoresolve)
+    {
+      return;
+    }
+
+  std::map<LPDISPATCH, Mail *>::iterator it;
+  for (it = s_mail_map.begin(); it != s_mail_map.end(); ++it)
+    {
+      if (it->second->needs_crypto ())
+        {
+          it->second->locate_keys ();
+        }
+    }
+}
+
+int
+Mail::remove_our_attachments ()
+{
+  LPDISPATCH attachments = get_oom_object (m_mailitem, "Attachments");
+  if (!attachments)
+    {
+      TRACEPOINT;
+      return 0;
+    }
+  int count = get_oom_int (attachments, "Count");
+  LPDISPATCH to_delete[count];
+  int del_cnt = 0;
+  for (int i = 1; i <= count; i++)
+    {
+      auto item_str = std::string("Item(") + std::to_string (i) + ")";
+      LPDISPATCH attachment = get_oom_object (attachments, item_str.c_str());
+      if (!attachment)
+        {
+          TRACEPOINT;
+          continue;
+        }
+
+      attachtype_t att_type;
+      if (get_pa_int (attachment, GPGOL_ATTACHTYPE_DASL, (int*) &att_type))
+        {
+          /* Not our attachment. */
+          gpgol_release (attachment);
+          continue;
+        }
+
+      if (att_type == ATTACHTYPE_PGPBODY || att_type == ATTACHTYPE_MOSS ||
+          att_type == ATTACHTYPE_MOSSTEMPL)
+        {
+          /* One of ours to delete. */
+          to_delete[del_cnt++] = attachment;
+          /* Dont' release yet */
+          continue;
+        }
+      gpgol_release (attachment);
+    }
+  gpgol_release (attachments);
+
+  int ret = 0;
+
+  for (int i = 0; i < del_cnt; i++)
+    {
+      LPDISPATCH attachment = to_delete[i];
+
+      /* Delete the attachments that are marked to delete */
+      if (invoke_oom_method (attachment, "Delete", NULL))
+        {
+          log_error ("%s:%s: Error: deleting attachment %i",
+                     SRCNAME, __func__, i);
+          ret = -1;
+        }
+      gpgol_release (attachment);
+    }
+  return ret;
+}
+
+/* We are very verbose because if we fail it might mean
+   that we have leaked plaintext -> critical. */
+bool
+Mail::has_crypted_or_empty_body ()
+{
+  const auto pair = has_crypt_or_empty_body_oom (this);
+
+  if (pair.first /* encrypted marker */)
+    {
+      log_debug ("%s:%s: Crypt Marker detected in OOM body. Return true %p.",
+                 SRCNAME, __func__, this);
+      return true;
+    }
+
+  if (!pair.second)
+    {
+      log_debug ("%s:%s: Unexpected content detected. Return false %p.",
+                 SRCNAME, __func__, this);
+      return false;
+    }
+
+  // Pair second == true (is empty) can happen on OOM error.
+  LPMESSAGE message = get_oom_base_message (m_mailitem);
+  if (!message && pair.second)
+    {
+      if (message)
+        {
+          gpgol_release (message);
+        }
+      return true;
+    }
+
+  size_t r_nbytes = 0;
+  char *mapi_body = mapi_get_body (message, &r_nbytes);
+  gpgol_release (message);
+
+  if (!mapi_body || !r_nbytes)
+    {
+      // Body or bytes are null. we are empty.
+      xfree (mapi_body);
+      log_debug ("%s:%s: MAPI error or empty message. Return true. %p.",
+                 SRCNAME, __func__, this);
+      return true;
+    }
+  if (r_nbytes > 10 && !strncmp (mapi_body, "-----BEGIN", 10))
+    {
+      // Body is crypt.
+      log_debug ("%s:%s: MAPI Crypt marker detected. Return true. %p.",
+                 SRCNAME, __func__, this);
+      xfree (mapi_body);
+      return true;
+    }
+
+  xfree (mapi_body);
+
+  log_debug ("%s:%s: Found mapi body. Return false. %p.",
+             SRCNAME, __func__, this);
+
+  return false;
+}
