@@ -32,6 +32,7 @@
 #include <future>
 
 class ParseController;
+class CryptController;
 
 /** @brief Data wrapper around a mailitem.
  *
@@ -47,6 +48,18 @@ class ParseController;
 class Mail
 {
 public:
+  enum CryptState
+    {
+      NoCryptMail,
+      NeedsFirstAfterWrite,
+      NeedsActualCrypt,
+      NeedsUpdateInOOM,
+      NeedsSecondAfterWrite,
+      NeedsUpdateInMAPI,
+      WantsSendInline,
+      WantsSendMIME,
+    };
+
   /** @brief Construct a mail object for the item.
     *
     * This also installs the event sink for this item.
@@ -73,6 +86,14 @@ public:
   */
   static Mail* get_mail_for_uuid (const char *uuid);
 
+  /** @brief Get the last created mail.
+
+    @returns A reference to the last created mail or null.
+  */
+  static Mail* get_last_mail ();
+
+  static void invalidate_last_mail ();
+
   /** @brief looks for existing Mail objects.
 
     @returns A reference to an existing mailitem or NULL in case none
@@ -114,6 +135,18 @@ public:
     */
   static int close_all_mails ();
 
+  /** @brief locate recipients for all crypto mails
+    *
+    * To avoid lookups of recipients for non crypto mails we only
+    * locate keys when a crypto action is already selected.
+    *
+    * As the user can do this after recipients were added but
+    * we don't know for which mail the crypt button was triggered.
+    * we march over all mails and if they are crypto mails we check
+    * that the recipents were located.
+    */
+  static void locate_all_crypto_recipients ();
+
   /** @brief Reference to the mailitem. Do not Release! */
   LPDISPATCH item () { return m_mailitem; }
 
@@ -134,13 +167,13 @@ public:
    * @returns 0 on success. */
   int decrypt_verify ();
 
-  /** @brief do crypto operations as selected by the user.
+  /** @brief start crypto operations as selected by the user.
    *
    * Initiates the crypto operations according to the gpgol
    * draft info flags.
    *
    * @returns 0 on success. */
-  int encrypt_sign ();
+  int encrypt_sign_start ();
 
   /** @brief Necessary crypto operations were completed successfully. */
   bool crypto_successful () { return !needs_crypto() || m_crypt_successful; }
@@ -183,9 +216,16 @@ public:
    * If the sender address has not been set through update_sender this
    * calls update_sender before returning the sender.
    *
-   * @returns A reference to the utf8 sender address. Or NULL. */
+   * @returns A reference to the utf8 sender address. Or an empty string. */
   std::string get_sender ();
 
+  /** @brief get sender SMTP address (UTF-8 encoded).
+   *
+   * Like get_sender but ensures not to touch oom or mapi
+   *
+   * @returns A reference to the utf8 sender address. Or an empty string. */
+  std::string get_cached_sender ();
+
   /** @brief get the subject string (UTF-8 encoded).
     *
     * @returns the subject or an empty string. */
@@ -226,6 +266,10 @@ public:
     only valid while the actual parsing happens. */
   std::shared_ptr<ParseController> parser () { return m_parser; }
 
+  /** @brief get the associated cryptcontroller.
+    only valid while the crypting happens. */
+  std::shared_ptr<CryptController> crypter () { return m_crypter; }
+
   /** To be called from outside once the paser was done.
    In Qt this would be a slot that is called once it is finished
    we hack around that a bit by calling it from our windowmessages
@@ -339,7 +383,10 @@ public:
   int check_attachments () const;
 
   /** Check if the mail should be encrypted "inline" */
-  bool should_inline_crypt () const {return m_do_inline;}
+  bool do_pgp_inline () const {return m_do_inline;}
+
+  /** Check if the mail should be encrypted "inline" */
+  void set_do_pgp_inline (bool value) {m_do_inline = value;}
 
   /** Append data to a cached inline body. Helper to do this
      on MAPI level and later add it through OOM */
@@ -347,9 +394,99 @@ public:
 
   /** Set the inline body as OOM body property. */
   int inline_body_to_body ();
+
+  /** Get the crypt state */
+  CryptState crypt_state () const {return m_crypt_state;}
+
+  /** Set the crypt state */
+  void set_crypt_state (CryptState state) {m_crypt_state = state;}
+
+  /** Update MAPI data after encryption. */
+  void update_crypt_mapi ();
+
+  /** Update OOM data after encryption.
+
+    Checks for plain text leaks and
+    does not advance crypt state if body can't be cleaned.
+  */
+  void update_crypt_oom ();
+
+  /** Enable / Disable the window of this mail.
+
+    When value is false the active window will
+    be disabled and the handle stored for a later
+    enable. */
+  void set_window_enabled (bool value);
+
+  /** Determine if the mail is an inline response.
+
+    Call check_inline_response first to update the state
+    from the OOM.
+
+    We need synchronous encryption for inline responses. */
+  bool async_crypt_disabled () { return m_async_crypt_disabled; }
+
+  /** Check through OOM if the current mail is an inline
+    response.
+
+    Caches the state which can then be queried through
+    async_crypt_disabled
+  */
+  bool check_inline_response ();
+
+  /** Get the window for the mail. Caution! This is only
+    really valid in the time that the window is disabled.
+    Use with care and can be null or invalid.
+  */
+  HWND get_window () { return m_window; }
+
+  /** Cleanup any attached crypter object. Useful
+    on error. */
+  void reset_crypter () { m_crypter = nullptr; }
+
+  /** Set special crypto mime data that should be used as the
+    mime structure when sending. */
+  void set_override_mime_data (const std::string &data) {m_mime_data = data;}
+
+  /** Get the mime data that should be used when sending. */
+  std::string get_override_mime_data () const { return m_mime_data; }
+
+  /** Set if this is a forward of a crypto mail. */
+  void set_is_forwarded_crypto_mail (bool value) { m_is_forwarded_crypto_mail = value; }
+  bool is_forwarded_crypto_mail () { return m_is_forwarded_crypto_mail; }
+
+  /** Remove the hidden GpgOL attachments. This is needed when forwarding
+    without encryption so that our attachments are not included in the forward.
+    Returns 0 on success. Works in OOM. */
+  int remove_our_attachments ();
+
+  /** Check both OOM and MAPI if the body is either empty or
+    encrypted. Won't abort on OOM or MAPI errors, so it can be
+    used in both states. But will return false if a body
+    was detected or in the OOM the MAPI Base Message. This
+    is intended as a saveguard before sending a mail.
+
+    This function should not be used to detected the necessity
+    of encryption and is only an extra check to catch unexpected
+    errors.
+    */
+  bool has_crypted_or_empty_body ();
+
+  void update_body ();
+
+  /** Set if this mail looks like the send again of a crypto mail.
+      This will mean that after it is decrypted it is treated
+      like an unencrypted mail so that it can be encrypted again
+      or sent unencrypted.
+      */
+  void set_is_send_again (bool value) { m_is_send_again = value; }
+
+
+  /* Attachment removal state variables. */
+  bool attachment_remove_warning_disabled () { return m_disable_att_remove_warning; }
+
 private:
   void update_categories ();
-  void update_body ();
   void update_sigstate ();
 
   LPDISPATCH m_mailitem;
@@ -373,6 +510,7 @@ private:
   char **m_cached_recipients;
   msgtype_t m_type; /* Our messagetype as set in mapi */
   std::shared_ptr <ParseController> m_parser;
+  std::shared_ptr <CryptController> m_crypter;
   GpgME::VerificationResult m_verify_result;
   GpgME::DecryptionResult m_decrypt_result;
   GpgME::Signature m_sig;
@@ -382,5 +520,12 @@ private:
   bool m_do_inline;
   bool m_is_gsuite; /* Are we on a gsuite account */
   std::string m_inline_body;
+  CryptState m_crypt_state;
+  HWND m_window;
+  bool m_async_crypt_disabled;
+  std::string m_mime_data;
+  bool m_is_forwarded_crypto_mail; /* Is this a forward of a crypto mail */
+  bool m_is_send_again; /* Is this a send again of a crypto mail */
+  bool m_disable_att_remove_warning; /* Should not warn about attachment removal. */
 };
 #endif // MAIL_H
