@@ -44,6 +44,8 @@ const wchar_t *prop_blacklist[] = {
   L"UnRead",
   L"OutlookVersion",
   L"OutlookInternalVersion",
+  L"ReceivedTime",
+  L"InternetCodepage",
   NULL };
 
 typedef enum
@@ -102,6 +104,7 @@ MailItemEvents::~MailItemEvents()
 }
 
 static bool propchangeWarnShown = false;
+static bool attachRemoveWarnShown = false;
 
 static DWORD WINAPI
 do_delayed_locate (LPVOID arg)
@@ -269,6 +272,21 @@ EVENT_SINK_INVOKE(MailItemEvents)
           log_oom ("%s:%s: Message %p propchange: %ls.",
                    SRCNAME, __func__, m_object, prop_name);
 
+          if (!wcscmp (prop_name, L"SendUsingAccount"))
+            {
+              bool sent = get_oom_bool (m_object, "Sent");
+              if (sent)
+                {
+                  log_debug ("%s:%s: Ignoring SendUsingAccount change for sent %p ",
+                             SRCNAME, __func__, m_object);
+                  return S_OK;
+                }
+              log_debug ("%s:%s: Message %p looks like send again.",
+                        SRCNAME, __func__, m_object);
+              m_mail->set_is_send_again (true);
+              return S_OK;
+            }
+
           /* We have tried several scenarios to handle propery changes.
              Only save the property in MAPI and call MAPI SaveChanges
              worked and did not leak plaintext but this caused outlook
@@ -335,7 +353,7 @@ EVENT_SINK_INVOKE(MailItemEvents)
            */
           log_oom_extra ("%s:%s: Send : %p",
                          SRCNAME, __func__, m_mail);
-          if (!m_mail->needs_crypto ())
+          if (!m_mail->needs_crypto () && m_mail->crypt_state () == Mail::NoCryptMail)
             {
              log_debug ("%s:%s: No crypto neccessary. Passing send for %p obj %p",
                         SRCNAME, __func__, m_mail, m_object);
@@ -348,107 +366,164 @@ EVENT_SINK_INVOKE(MailItemEvents)
                         SRCNAME, __func__);
              break;
            }
-          m_mail->update_oom_data ();
-          m_mail->set_needs_encrypt (true);
 
-          invoke_oom_method (m_object, "Save", NULL);
-          if (m_mail->crypto_successful ())
+          if (m_mail->crypt_state () == Mail::NoCryptMail &&
+              m_mail->needs_crypto ())
             {
-              /* Fishy behavior catcher: Sometimes the save does not clean
-                 out the body. Weird. Happened at least for one user.
-                 The following code checks for plain text leaks and
-                 prevents send in case the body can't be wiped. */
-              if (!m_mail->get_body().empty() || !m_mail->get_html_body().empty())
+              // First contact with a mail to encrypt update
+              // state and oom data.
+              m_mail->update_oom_data ();
+              m_mail->set_crypt_state (Mail::NeedsFirstAfterWrite);
+
+              // Check inline response state before the write.
+              m_mail->check_inline_response ();
+
+              log_debug ("%s:%s: Send event for crypto mail %p saving and starting.",
+                         SRCNAME, __func__, m_mail);
+              // Save the Mail
+              invoke_oom_method (m_object, "Save", NULL);
+
+              if (!m_mail->async_crypt_disabled ())
                 {
-                  log_debug ("%s:%s: Body found after encryption %p.",
-                            SRCNAME, __func__, m_object);
-                  if (m_mail->should_inline_crypt ())
+                  // The afterwrite in the save should have triggered
+                  // the encryption. We cancel send for our asyncness.
+                  // Cancel send
+                  *(parms->rgvarg[0].pboolVal) = VARIANT_TRUE;
+                  break;
+                }
+              else
+                {
+                  if (m_mail->crypt_state () == Mail::NoCryptMail)
                     {
-                      if (m_mail->inline_body_to_body ())
-                        {
-                          log_debug ("%s:%s: Inline body to body failed %p.",
-                                    SRCNAME, __func__, m_object);
-                        }
-                    }
-                  const auto body = m_mail->get_body();
-                  if (body.size() > 10 && !strncmp (body.c_str(), "-----BEGIN", 10))
-                    {
-                      log_debug ("%s:%s: Looks like inline body. You can pass %p.",
-                                SRCNAME, __func__, m_object);
-                      break;
-                    }
-
-                  if (m_mail->wipe (true))
-                    {
-                      log_debug ("%s:%s: Cancel send for %p.",
-                                SRCNAME, __func__, m_object);
-                      wchar_t *title = utf8_to_wchar (_("GpgOL: Encryption not possible!"));
-                      wchar_t *msg = utf8_to_wchar (_(
-                      "Outlook returned an error when trying to send the encrypted mail.\n\n"
-                      "Please restart Outlook and try again.\n\n"
-                      "If it still fails consider using an encrypted attachment or\n"
-                      "switching to PGP/Inline in GpgOL's options."));
-                      MessageBoxW (get_active_hwnd(), msg, title,
-                                   MB_ICONERROR | MB_OK);
-                      xfree (msg);
-                      xfree (title);
+                      // Crypto failed or was canceled
+                      log_debug ("%s:%s: Message %p mail %p cancelling send - "
+                                 "Crypto failed or canceled.",
+                                 SRCNAME, __func__, m_object, m_mail);
                       *(parms->rgvarg[0].pboolVal) = VARIANT_TRUE;
-                      return S_OK;
-                    }
-                    {
-                      /* Now we adress T3656 if Outlooks internal S/MIME is somehow
-                       * mixed in (even if it is enabled and then disabled) it might
-                       * cause strange behavior in that it sends the plain message
-                       * and not the encrypted message. Tests have shown that we can
-                       * bypass that by calling submit message on our base
-                       * message.
-                       *
-                       * We do this conditionally as our other way of using OOM
-                       * to send is proven to work and we don't want to mess
-                       * with it.
-                       */
-                      // Get the Message class.
-                      HRESULT hr;
-                      LPSPropValue propval = NULL;
-
-                      // It's important we use the _not_ base message here.
-                      LPMESSAGE message = get_oom_message (m_object);
-                      hr = HrGetOneProp ((LPMAPIPROP)message, PR_MESSAGE_CLASS_A, &propval);
-                      gpgol_release (message);
-                      if (FAILED (hr))
-                        {
-                          log_error ("%s:%s: HrGetOneProp() failed: hr=%#lx\n",
-                                     SRCNAME, __func__, hr);
-                          gpgol_release (message);
-                          break;
-                        }
-                      if (propval->Value.lpszA && !strstr (propval->Value.lpszA, "GpgOL"))
-                        {
-                          // Does not have a message class by us.
-                          log_debug ("%s:%s: Message %p - No GpgOL Message class after encryption.",
-                                    SRCNAME, __func__, m_object);
-                          log_debug ("%s:%s: Message %p - Activating T3656 Workaround",
-                                    SRCNAME, __func__, m_object);
-                          message = get_oom_base_message (m_object);
-                          // It's important we use the _base_ message here.
-                          mapi_save_changes (message, KEEP_OPEN_READWRITE | FORCE_SAVE);
-                          message->SubmitMessage(0);
-                          gpgol_release (message);
-
-                          // Cancel send
-                          *(parms->rgvarg[0].pboolVal) = VARIANT_TRUE;
-
-                          // Close the composer and trigger unloads
-                          CloseHandle(CreateThread (NULL, 0, close_mail, (LPVOID) m_mail, 0,
-                                                    NULL));
-                        }
-                      MAPIFreeBuffer (propval);
-                    }
-
-                  if (*(parms->rgvarg[0].pboolVal) == VARIANT_TRUE)
-                    {
                       break;
                     }
+                  // For inline response we can't trigger send programatically
+                  // so we do the encryption in sync.
+                  if (m_mail->crypt_state () == Mail::NeedsUpdateInOOM)
+                    {
+                      m_mail->update_crypt_oom ();
+                    }
+                  if (m_mail->crypt_state () == Mail::NeedsSecondAfterWrite)
+                    {
+                      m_mail->set_crypt_state (Mail::WantsSendMIME);
+                    }
+                  if (m_mail->do_pgp_inline () && m_mail->crypt_state () != Mail::WantsSendInline)
+                    {
+                      log_debug ("%s:%s: Message %p mail %p cancelling send - "
+                                 "Invalid state.",
+                                 SRCNAME, __func__, m_object, m_mail);
+                      gpgol_bug (m_mail->get_window (),
+                                 ERR_INLINE_BODY_INV_STATE);
+                      *(parms->rgvarg[0].pboolVal) = VARIANT_TRUE;
+                      break;
+                    }
+                }
+            }
+
+          if (m_mail->crypt_state () == Mail::WantsSendInline)
+            {
+              if (!m_mail->has_crypted_or_empty_body())
+                {
+                  log_debug ("%s:%s: Message %p mail %p cancelling send - "
+                             "not encrypted or not empty body detected.",
+                             SRCNAME, __func__, m_object, m_mail);
+                  gpgol_bug (m_mail->get_window (),
+                             ERR_WANTS_SEND_INLINE_BODY);
+                  m_mail->set_crypt_state (Mail::NoCryptMail);
+                  *(parms->rgvarg[0].pboolVal) = VARIANT_TRUE;
+                  break;
+                }
+              log_debug ("%s:%s: Passing send event for no-mime message %p.",
+                         SRCNAME, __func__, m_object);
+              break;
+            }
+
+          if (m_mail->crypt_state () == Mail::WantsSendMIME)
+            {
+              if (!m_mail->has_crypted_or_empty_body())
+                {
+/* The safety checks here trigger too often. Somehow for some
+   users the body is not empty after the encryption but when
+   it is sent it is still sent with the crypto content because
+   the encrypted MIME Structure is used because it is
+   correct in MAPI land.
+
+   For safety reasons enabling the checks might be better but
+   until we figure out why for some users the body replacement
+   does not work we have to disable them. Otherwise GpgOL
+   is unusuable for such users. GnuPG-Bug-Id: T3875
+*/
+#define DISABLE_SAFTEY_CHECKS
+#ifndef DISABLE_SAFTEY_CHECKS
+                  gpgol_bug (m_mail->get_window (),
+                             ERR_WANTS_SEND_MIME_BODY);
+                  log_debug ("%s:%s: Message %p mail %p cancelling send mime - "
+                             "not encrypted or not empty body detected.",
+                             SRCNAME, __func__, m_object, m_mail);
+                  m_mail->set_crypt_state (Mail::NoCryptMail);
+                  *(parms->rgvarg[0].pboolVal) = VARIANT_TRUE;
+                  break;
+#else
+                  log_debug ("%s:%s: Message %p mail %p - "
+                             "not encrypted or not empty body detected - MIME.",
+                             SRCNAME, __func__, m_object, m_mail);
+#endif
+                }
+              /* Now we adress T3656 if Outlooks internal S/MIME is somehow
+               * mixed in (even if it is enabled and then disabled) it might
+               * cause strange behavior in that it sends the plain message
+               * and not the encrypted message. Tests have shown that we can
+               * bypass that by calling submit message on our base
+               * message.
+               *
+               * We do this conditionally as our other way of using OOM
+               * to send is proven to work and we don't want to mess
+               * with it.
+               */
+              // Get the Message class.
+              HRESULT hr;
+              LPSPropValue propval = NULL;
+
+              // It's important we use the _not_ base message here.
+              LPMESSAGE message = get_oom_message (m_object);
+              hr = HrGetOneProp ((LPMAPIPROP)message, PR_MESSAGE_CLASS_A, &propval);
+              gpgol_release (message);
+              if (FAILED (hr))
+                {
+                  log_error ("%s:%s: HrGetOneProp() failed: hr=%#lx\n",
+                             SRCNAME, __func__, hr);
+                  gpgol_release (message);
+                  break;
+                }
+              if (propval->Value.lpszA && !strstr (propval->Value.lpszA, "GpgOL"))
+                {
+                  // Does not have a message class by us.
+                  log_debug ("%s:%s: Message %p - No GpgOL Message class after encryption. cls is: '%s'",
+                             SRCNAME, __func__, m_object, propval->Value.lpszA);
+                  log_debug ("%s:%s: Message %p - Activating T3656 Workaround",
+                             SRCNAME, __func__, m_object);
+                  message = get_oom_base_message (m_object);
+                  // It's important we use the _base_ message here.
+                  mapi_save_changes (message, KEEP_OPEN_READWRITE | FORCE_SAVE);
+                  message->SubmitMessage(0);
+                  gpgol_release (message);
+
+                  // Cancel send
+                  *(parms->rgvarg[0].pboolVal) = VARIANT_TRUE;
+
+                  // Close the composer and trigger unloads
+                  CloseHandle(CreateThread (NULL, 0, close_mail, (LPVOID) m_mail, 0,
+                                            NULL));
+                }
+              MAPIFreeBuffer (propval);
+              if (*(parms->rgvarg[0].pboolVal) == VARIANT_TRUE)
+                {
+                  break;
                 }
               log_debug ("%s:%s: Passing send event for message %p.",
                          SRCNAME, __func__, m_object);
@@ -456,7 +531,8 @@ EVENT_SINK_INVOKE(MailItemEvents)
             }
           else
             {
-              log_debug ("%s:%s: Message %p cancelling send - crypto failed.",
+              log_debug ("%s:%s: Message %p cancelling send - "
+                         "crypto or second save failed.",
                          SRCNAME, __func__, m_object);
               *(parms->rgvarg[0].pboolVal) = VARIANT_TRUE;
             }
@@ -481,6 +557,46 @@ EVENT_SINK_INVOKE(MailItemEvents)
 
           if (m_mail->is_crypto_mail () && !m_mail->needs_save ())
             {
+              Mail *last_mail = Mail::get_last_mail ();
+              if (Mail::is_valid_ptr (last_mail))
+                {
+                  /* We want to identify here if there was a mail created that
+                     should receive the contents of this mail. For this we check
+                     for a write in the same loop as a mail creation.
+                     Now when switching from one mail to another this is also what
+                     happens. The new mail is loaded and the old mail is written.
+                     To distinguish the two we check that the new mail does not have
+                     an entryID, a Subject and No Size. Maybe just size or entryID
+                     would be enough but better save then sorry.
+
+                     Security consideration: Worst case we pass the write here but
+                     an unload follows before we get the scheduled revert. This
+                     would leak plaintext. But does not happen in our tests.
+
+                     Similarly if we crash or Outlook is closed before we see this
+                     revert. But as we immediately revert after the write this should
+                     also not happen. */
+                  const std::string lastSubject = last_mail->get_subject ();
+                  char *lastEntryID = get_oom_string (last_mail->item (), "EntryID");
+                  int lastSize = get_oom_int (last_mail->item (), "Size");
+                  std::string lastEntryStr;
+                  if (lastEntryID)
+                    {
+                      lastEntryStr = lastEntryID;
+                      xfree (lastEntryID);
+                    }
+
+                  if (!lastSize && !lastEntryStr.size () && !lastSubject.size ())
+                    {
+                      log_debug ("%s:%s: Write in the same loop as empty load."
+                                 " Pass but schedule revert.",
+                                 SRCNAME, __func__);
+
+                      Mail::invalidate_last_mail ();
+                      do_in_ui_thread_async (REVERT_MAIL, m_mail);
+                      return S_OK;
+                    }
+                }
               /* We cancel the write event to stop outlook from excessively
                  syncing our changes.
                  if smime support is disabled and we still have an smime
@@ -504,8 +620,27 @@ EVENT_SINK_INVOKE(MailItemEvents)
               *(parms->rgvarg[0].pboolVal) = VARIANT_TRUE;
             }
 
-          log_debug ("%s:%s: Passing write event. Subject: %s Body: %s MsgCls: %s",
-                     SRCNAME, __func__, m_mail->get_subject().c_str(), m_mail->get_body().c_str(), get_oom_string (m_object, "MessageClass"));
+          if (!m_mail->is_crypto_mail () && m_mail->is_forwarded_crypto_mail () &&
+              !m_mail->needs_crypto () && m_mail->crypt_state () == Mail::NoCryptMail)
+            {
+              /* We are sure now that while this is a forward of an encrypted
+               * mail that the forward should not be signed or encrypted. So
+               * it's not constructed by us. We need to remove our attachments
+               * though so that they are not included in the forward. */
+              log_debug ("%s:%s: Writing unencrypted forward of crypt mail. "
+                         "Removing attachments. mail: %p item: %p",
+                         SRCNAME, __func__, m_mail, m_object);
+              if (m_mail->remove_our_attachments ())
+                {
+                  // Worst case we forward some encrypted data here not
+                  // a security problem, so let it pass.
+                  log_error ("%s:%s: Failed to remove our attachments.",
+                             SRCNAME, __func__);
+                }
+              /* Remove marker because we did this now. */
+              m_mail->set_is_forwarded_crypto_mail (false);
+            }
+
           log_debug ("%s:%s: Passing write event.",
                      SRCNAME, __func__);
           m_mail->set_needs_save (false);
@@ -515,9 +650,22 @@ EVENT_SINK_INVOKE(MailItemEvents)
         {
           log_oom_extra ("%s:%s: AfterWrite : %p",
                          SRCNAME, __func__, m_mail);
-          if (m_mail->needs_encrypt ())
+          if (m_mail->crypt_state () == Mail::NeedsFirstAfterWrite)
             {
-              m_mail->encrypt_sign ();
+              /* Seen the first after write. Advance the state */
+              m_mail->set_crypt_state (Mail::NeedsActualCrypt);
+              if (m_mail->encrypt_sign_start ())
+                {
+                  log_debug ("%s:%s: Encrypt sign start failes.",
+                             SRCNAME, __func__);
+                  m_mail->set_crypt_state (Mail::NoCryptMail);
+                }
+              return S_OK;
+            }
+          if (m_mail->crypt_state () == Mail::NeedsSecondAfterWrite)
+            {
+              m_mail->set_crypt_state (Mail::NeedsUpdateInMAPI);
+              m_mail->update_crypt_mapi ();
               return S_OK;
             }
           break;
@@ -567,9 +715,46 @@ EVENT_SINK_INVOKE(MailItemEvents)
           log_debug ("%s:%s: Removing Mail for message: %p.",
                      SRCNAME, __func__, m_object);
           delete m_mail;
+          log_oom_extra ("%s:%s: deletion done",
+                         SRCNAME, __func__);
           return S_OK;
         }
       case Forward:
+        {
+          log_oom_extra ("%s:%s: Forward: %p",
+                         SRCNAME, __func__, m_mail);
+          if (!m_mail->is_crypto_mail ())
+            {
+              /* Non crypto mails do not interest us.*/
+              break;
+            }
+          Mail *last_mail = Mail::get_last_mail ();
+          if (Mail::is_valid_ptr (last_mail))
+            {
+              /* We want to identify here if there was a mail created that
+                 should receive the contents of this mail. For this we check
+                 for a forward in the same loop as a mail creation.
+              */
+              char *lastEntryID = get_oom_string (last_mail->item (), "EntryID");
+              int lastSize = get_oom_int (last_mail->item (), "Size");
+              std::string lastEntryStr;
+              if (lastEntryID)
+                {
+                  lastEntryStr = lastEntryID;
+                  xfree (lastEntryID);
+                }
+
+              if (!lastSize && !lastEntryStr.size ())
+                {
+                  log_debug ("%s:%s: Forward in the same loop as empty load."
+                             " Marking %p (item %p) as forwarded.",
+                             SRCNAME, __func__, last_mail, last_mail->item ());
+
+                  last_mail->set_is_forwarded_crypto_mail (true);
+                }
+            }
+        }
+      /* Fallthrough */
       case Reply:
       case ReplyAll:
         {
@@ -604,6 +789,23 @@ EVENT_SINK_INVOKE(MailItemEvents)
           gpgol_release (msg);
           break;
         }
+      case AttachmentRemove:
+        {
+          log_oom_extra ("%s:%s: AttachmentRemove: %p",
+                         SRCNAME, __func__, m_mail);
+          if (!m_mail->is_crypto_mail () || attachRemoveWarnShown ||
+              m_mail->attachment_remove_warning_disabled ())
+            {
+              return S_OK;
+            }
+          gpgol_message_box (get_active_hwnd (),
+                             _("Attachments are part of the crypto message.\nThey "
+                               "can't be permanently removed and will be shown again the next "
+                               "time this message is opened."),
+                             _("Sorry, that's not possible, yet"), MB_OK);
+          attachRemoveWarnShown = true;
+          return S_OK;
+        }
 
       default:
         log_oom_extra ("%s:%s: Message:%p Unhandled Event: %lx \n",
