@@ -152,8 +152,9 @@ check_protocol (protocol_t protocol)
    to is stored at STREAM and the attachment object itself is
    returned.  The caller needs to call SaveChanges.  Returns NULL on
    failure in which case STREAM will be set to NULL.  */
-static LPATTACH
-create_mapi_attachment (LPMESSAGE message, sink_t sink)
+LPATTACH
+create_mapi_attachment (LPMESSAGE message, sink_t sink,
+                        const char *overrideMimeTag)
 {
   HRESULT hr;
   ULONG pos;
@@ -217,7 +218,13 @@ create_mapi_attachment (LPMESSAGE message, sink_t sink)
   if (!hr)
     {
       prop.ulPropTag = PR_ATTACH_MIME_TAG_A;
-      prop.Value.lpszA = strdup("multipart/signed");
+      prop.Value.lpszA = overrideMimeTag ? strdup (overrideMimeTag) :
+                         strdup ("multipart/signed");
+      if (overrideMimeTag)
+        {
+          log_debug ("%s:%s: using override mimetag: %s\n",
+                     SRCNAME, __func__, overrideMimeTag);
+        }
       hr = HrSetOneProp ((LPMAPIPROP)att, &prop);
       xfree (prop.Value.lpszA);
     }
@@ -273,7 +280,7 @@ write_buffer_for_cb (void *opaque, const void *data, size_t datalen)
 
 /* Write the string TEXT to the IStream STREAM.  Returns 0 on sucsess,
    prints an error message and returns -1 on error.  */
-static int
+int
 write_string (sink_t sink, const char *text)
 {
   return write_buffer (sink, text, strlen (text));
@@ -304,7 +311,7 @@ write_multistring (sink_t sink, const char *text1, ...)
 
 /* Helper to write a boundary to the output sink.  The leading LF
    will be written as well.  */
-static int
+int
 write_boundary (sink_t sink, const char *boundary, int lastone)
 {
   int rc = write_string (sink, "\r\n--");
@@ -318,7 +325,7 @@ write_boundary (sink_t sink, const char *boundary, int lastone)
 
 /* Write DATALEN bytes of DATA to SINK in base64 encoding.  This
    creates a complete Base64 chunk including the trailing fillers.  */
-static int
+int
 write_b64 (sink_t sink, const void *data, size_t datalen)
 {
   int rc;
@@ -1084,7 +1091,7 @@ delete_all_attachments (LPMESSAGE message, mapi_attach_item_t *table)
    SINK needs to be passed as well and will also be closed.  Note that
    the address of ATTACH is expected so that the fucntion can set it
    to NULL. */
-static int
+int
 close_mapi_attachment (LPATTACH *attach, sink_t sink)
 {
   HRESULT hr;
@@ -1121,7 +1128,7 @@ close_mapi_attachment (LPATTACH *attach, sink_t sink)
    SINK needs to be passed as well and will also be closed.  Note that
    the address of ATTACH is expected so that the fucntion can set it
    to NULL. */
-static void
+void
 cancel_mapi_attachment (LPATTACH *attach, sink_t sink)
 {
   LPSTREAM stream = sink ? (LPSTREAM) sink->cb_data : NULL;
@@ -1143,9 +1150,9 @@ cancel_mapi_attachment (LPATTACH *attach, sink_t sink)
 
 
 /* Do the final processing for a message. */
-static int
+int
 finalize_message (LPMESSAGE message, mapi_attach_item_t *att_table,
-                  protocol_t protocol, int encrypt, bool is_inline = false)
+                  protocol_t protocol, int encrypt, bool is_inline)
 {
   HRESULT hr = 0;
   SPropValue prop;
@@ -1298,7 +1305,7 @@ collect_signature (void *opaque, const void *data, size_t datalen)
 /* Helper to create the signing header.  This includes enough space
    for later fixup of the micalg parameter.  The MIME version is only
    written if FIRST is set.  */
-static void
+void
 create_top_signing_header (char *buffer, size_t buflen, protocol_t protocol,
                            int first, const char *boundary, const char *micalg)
 {
@@ -1403,7 +1410,7 @@ add_body (Mail *mail, const char *boundary, sink_t sink,
 }
 
 /* Add the body and attachments. Does multipart handling. */
-static int
+int
 add_body_and_attachments (sink_t sink, LPMESSAGE message,
                           mapi_attach_item_t *att_table, Mail *mail,
                           const char *body, int n_att_usable)
@@ -1883,9 +1890,9 @@ sink_encryption_write_b64 (sink_t encsink, const void *data, size_t datalen)
 /* Helper from mime_encrypt.  BOUNDARY is a buffer of at least
    BOUNDARYSIZE+1 bytes which will be set on return from that
    function.  */
-static int
+int
 create_top_encryption_header (sink_t sink, protocol_t protocol, char *boundary,
-                              bool is_inline = false)
+                              bool is_inline, int exchange_major_version)
 {
   int rc;
 
@@ -1908,14 +1915,31 @@ create_top_encryption_header (sink_t sink, protocol_t protocol, char *boundary,
   else if (protocol == PROTOCOL_SMIME)
     {
       *boundary = 0;
-      rc = write_multistring (sink,
-                              "MIME-Version: 1.0\r\n"
-                              "Content-Type: application/pkcs7-mime;\r\n"
-                              "\tsmime-type=enveloped-data;\r\n"
-                              "\tname=\"smime.p7m\"\r\n"
-                              "Content-Transfer-Encoding: base64\r\n"
-                              "\r\n",
-                              NULL);
+      if (exchange_major_version >= 15)
+        {
+          /*
+             For S/MIME encrypted mails we do not use the S/MIME conversion
+             code anymore. With Exchange 2016 this no longer works. Instead
+             we set an override mime tag, the extended headers in OOM in
+             Mail::update_crypt_oom and let outlook convert the attachment
+             to base64.
+
+             A bit more details can be found in T3853 / T3884
+             */
+          rc = 0;
+        }
+      else
+        {
+          rc = write_multistring (sink,
+                                  "Content-Type: application/pkcs7-mime; "
+                                  "smime-type=enveloped-data;\r\n"
+                                  "\tname=\"smime.p7m\"\r\n"
+                                  "Content-Disposition: attachment; filename=\"smime.p7m\"\r\n"
+                                  "Content-Transfer-Encoding: base64\r\n"
+                                  "MIME-Version: 1.0\r\n"
+                                  "\r\n",
+                                  NULL);
+        }
     }
   else
     {
@@ -1972,7 +1996,7 @@ mime_encrypt (LPMESSAGE message, HWND hwnd,
   int n_att_usable;
   engine_filter_t filter = NULL;
   char *my_sender = NULL;
-  bool is_inline = mail && mail->should_inline_crypt ();
+  bool is_inline = mail && mail->do_pgp_inline ();
 
   memset (sink, 0, sizeof *sink);
   memset (encsink, 0, sizeof *encsink);
