@@ -73,6 +73,27 @@ int restore_msg_from_moss (LPMESSAGE message, LPDISPATCH moss_att,
 
 int count_usable_attachments (mapi_attach_item_t *table);
 
+int add_body_and_attachments (sink_t sink, LPMESSAGE message,
+                              mapi_attach_item_t *att_table, Mail *mail,
+                              const char *body, int n_att_usable);
+int create_top_encryption_header (sink_t sink, protocol_t protocol, char *boundary,
+                              bool is_inline = false, int exchange_major_version = -1);
+
+/* Helper to write a boundary to the output sink.  The leading LF
+   will be written as well.  */
+int write_boundary (sink_t sink, const char *boundary, int lastone);
+
+LPATTACH create_mapi_attachment (LPMESSAGE message, sink_t sink,
+                                 const char *overrideMimeTag = nullptr);
+int close_mapi_attachment (LPATTACH *attach, sink_t sink);
+int finalize_message (LPMESSAGE message, mapi_attach_item_t *att_table,
+                      protocol_t protocol, int encrypt, bool is_inline = false);
+void cancel_mapi_attachment (LPATTACH *attach, sink_t sink);
+void create_top_signing_header (char *buffer, size_t buflen, protocol_t protocol,
+                           int first, const char *boundary, const char *micalg);
+int write_string (sink_t sink, const char *text);
+int write_b64 (sink_t sink, const void *data, size_t datalen);
+
 #ifdef __cplusplus
 }
 #endif
