@@ -533,6 +533,40 @@ put_oom_string (LPDISPATCH pDisp, const char *name, const char *string)
   return 0;
 }
 
+/* Set the property NAME to DISP.  */
+int
+put_oom_disp (LPDISPATCH pDisp, const char *name, LPDISPATCH disp)
+{
+  HRESULT hr;
+  DISPID dispid_put = DISPID_PROPERTYPUT;
+  DISPID dispid;
+  DISPPARAMS dispparams;
+  VARIANT aVariant[1];
+  EXCEPINFO execpinfo;
+
+  init_excepinfo (&execpinfo);
+  dispid = lookup_oom_dispid (pDisp, name);
+  if (dispid == DISPID_UNKNOWN)
+    return -1;
+
+  dispparams.rgvarg = aVariant;
+  dispparams.rgvarg[0].vt = VT_DISPATCH;
+  dispparams.rgvarg[0].pdispVal = disp;
+  dispparams.cArgs = 1;
+  dispparams.rgdispidNamedArgs = &dispid_put;
+  dispparams.cNamedArgs = 1;
+  hr = pDisp->Invoke (dispid, IID_NULL, LOCALE_SYSTEM_DEFAULT,
+                      DISPATCH_PROPERTYPUTREF, &dispparams,
+                      NULL, &execpinfo, NULL);
+  if (hr != S_OK)
+    {
+      log_debug ("%s:%s: Putting '%s' failed: %#lx",
+                 SRCNAME, __func__, name, hr);
+      dump_excepinfo (execpinfo);
+      return -1;
+    }
+  return 0;
+}
 
 /* Get the boolean property NAME of the object PDISP.  Returns False if
    not found or if it is not a boolean property.  */
@@ -1749,8 +1783,16 @@ get_active_hwnd ()
   LPDISPATCH activeWindow = get_oom_object (app, "ActiveWindow");
   if (!activeWindow)
     {
-      TRACEPOINT;
-      return nullptr;
+      activeWindow = get_oom_object (app, "ActiveInspector");
+      if (!activeWindow)
+        {
+          activeWindow = get_oom_object (app, "ActiveExplorer");
+          if (!activeWindow)
+            {
+              TRACEPOINT;
+              return nullptr;
+            }
+        }
     }
 
   /* Both explorer and inspector have this. */
@@ -1769,3 +1811,249 @@ get_active_hwnd ()
 
   return hwnd;
 }
+
+LPDISPATCH
+create_mail ()
+{
+  LPDISPATCH app = GpgolAddin::get_instance ()->get_application ();
+
+  if (!app)
+    {
+      TRACEPOINT;
+      return nullptr;
+   }
+
+  VARIANT var;
+  VariantInit (&var);
+  VARIANT argvars[1];
+  DISPPARAMS args;
+  VariantInit (&argvars[0]);
+  argvars[0].vt = VT_I2;
+  argvars[0].intVal = 0;
+  args.cArgs = 1;
+  args.cNamedArgs = 0;
+  args.rgvarg = argvars;
+
+  LPDISPATCH ret = nullptr;
+
+  if (invoke_oom_method_with_parms (app, "CreateItem", &var, &args))
+    {
+      log_error ("%s:%s: Failed to create mailitem.",
+                 SRCNAME, __func__);
+      return ret;
+    }
+
+  ret = var.pdispVal;
+  return ret;
+}
+
+LPDISPATCH
+get_account_for_mail (const char *mbox)
+{
+  LPDISPATCH app = GpgolAddin::get_instance ()->get_application ();
+
+  if (!app)
+    {
+      TRACEPOINT;
+      return nullptr;
+   }
+
+  LPDISPATCH accounts = get_oom_object (app, "Session.Accounts");
+
+  if (!accounts)
+    {
+      TRACEPOINT;
+      return nullptr;
+    }
+
+  int count = get_oom_int (accounts, "Count");
+  for (int i = 1; i <= count; i++)
+    {
+      std::string item = std::string ("Item(") + std::to_string (i) + ")";
+
+      LPDISPATCH account = get_oom_object (accounts, item.c_str ());
+
+      if (!account)
+        {
+          TRACEPOINT;
+          continue;
+        }
+      char *smtpAddr = get_oom_string (account, "SmtpAddress");
+
+      if (!smtpAddr)
+        {
+          TRACEPOINT;
+          continue;
+        }
+      if (!stricmp (mbox, smtpAddr))
+        {
+          gpgol_release (accounts);
+          xfree (smtpAddr);
+          return account;
+        }
+      xfree (smtpAddr);
+    }
+  gpgol_release (accounts);
+
+  log_error ("%s:%s: Failed to find account for '%s'.",
+             SRCNAME, __func__, mbox);
+
+  return nullptr;
+}
+
+char *
+get_sender_SendUsingAccount (LPDISPATCH mailitem, bool *r_is_GSuite)
+{
+  LPDISPATCH sender = get_oom_object (mailitem, "SendUsingAccount");
+  if (!sender)
+    {
+      return nullptr;
+    }
+
+  char *buf = get_oom_string (sender, "SmtpAddress");
+  char *dispName = get_oom_string (sender, "DisplayName");
+  gpgol_release (sender);
+
+  /* Check for G Suite account */
+  if (dispName && !strcmp ("G Suite", dispName) && r_is_GSuite)
+    {
+      *r_is_GSuite = true;
+    }
+  xfree (dispName);
+  if (buf && strlen (buf))
+    {
+      log_debug ("%s:%s: found sender", SRCNAME, __func__);
+      return buf;
+    }
+  xfree (buf);
+  return nullptr;
+}
+
+char *
+get_sender_Sender (LPDISPATCH mailitem)
+{
+  LPDISPATCH sender = get_oom_object (mailitem, "Sender");
+  if (!sender)
+    {
+      return nullptr;
+    }
+  char *buf = get_pa_string (sender, PR_SMTP_ADDRESS_DASL);
+  gpgol_release (sender);
+  if (buf && strlen (buf))
+    {
+      log_debug ("%s:%s Sender fallback 2", SRCNAME, __func__);
+      return buf;
+    }
+  xfree (buf);
+  /* We have a sender object but not yet an smtp address likely
+     exchange. Try some more propertys of the message. */
+  buf = get_pa_string (mailitem, PR_TAG_SENDER_SMTP_ADDRESS);
+  if (buf && strlen (buf))
+    {
+      log_debug ("%s:%s Sender fallback 3", SRCNAME, __func__);
+      return buf;
+    }
+  xfree (buf);
+  buf = get_pa_string (mailitem, PR_TAG_RECEIVED_REPRESENTING_SMTP_ADDRESS);
+  if (buf && strlen (buf))
+    {
+      log_debug ("%s:%s Sender fallback 4", SRCNAME, __func__);
+      return buf;
+    }
+  xfree (buf);
+  return nullptr;
+}
+
+char *
+get_sender_CurrentUser (LPDISPATCH mailitem)
+{
+  LPDISPATCH sender = get_oom_object (mailitem,
+                                      "Session.CurrentUser");
+  if (!sender)
+    {
+      return nullptr;
+    }
+  char *buf = get_pa_string (sender, PR_SMTP_ADDRESS_DASL);
+  gpgol_release (sender);
+  if (buf && strlen (buf))
+    {
+      log_debug ("%s:%s Sender fallback 5", SRCNAME, __func__);
+      return buf;
+    }
+  xfree (buf);
+  return nullptr;
+}
+
+char *
+get_sender_SenderEMailAddress (LPDISPATCH mailitem)
+{
+
+  char *type = get_oom_string (mailitem, "SenderEmailType");
+  if (type && !strcmp ("SMTP", type))
+    {
+      char *senderMail = get_oom_string (mailitem, "SenderEmailAddress");
+      if (senderMail)
+        {
+          log_debug ("%s:%s: Sender found", SRCNAME, __func__);
+          xfree (type);
+          return senderMail;
+        }
+    }
+  xfree (type);
+  return nullptr;
+}
+
+char *
+get_inline_body ()
+{
+  LPDISPATCH app = GpgolAddin::get_instance ()->get_application ();
+  if (!app)
+    {
+      TRACEPOINT;
+      return nullptr;
+    }
+
+  LPDISPATCH explorer = get_oom_object (app, "ActiveExplorer");
+
+  if (!explorer)
+    {
+      TRACEPOINT;
+      return nullptr;
+    }
+
+  LPDISPATCH inlineResponse = get_oom_object (explorer, "ActiveInlineResponse");
+  gpgol_release (explorer);
+
+  if (!inlineResponse)
+    {
+      return nullptr;
+    }
+
+  char *body = get_oom_string (inlineResponse, "Body");
+  gpgol_release (inlineResponse);
+
+  return body;
+}
+
+int
+get_ex_major_version_for_addr (const char *mbox)
+{
+  LPDISPATCH account = get_account_for_mail (mbox);
+  if (!account)
+    {
+      TRACEPOINT;
+      return -1;
+    }
+
+  char *version_str = get_oom_string (account, "ExchangeMailboxServerVersion");
+  gpgol_release (account);
+
+  if (!version_str)
+    {
+      return -1;
+    }
+  long int version = strtol (version_str, nullptr, 10);
+  xfree (version_str);
+
+  return (int) version;
+}
