@@ -194,7 +194,15 @@ format_error(GpgME::DecryptionResult result, Protocol protocol)
   else
     {
       msg = _("Could not decrypt the data: ");
-      msg += result.error().asString();
+
+      if (result.isNull ())
+        {
+          msg += _("Failed to parse the mail.");
+        }
+      else
+        {
+          msg += result.error().asString();
+        }
     }
 
   if (gpgrt_asprintf (&buf, opt.prefer_html ? decrypt_template_html :
@@ -226,7 +234,9 @@ ParseController::parse()
 
   Data input (m_inputprovider);
 
-  if (input.type () == Data::Type::PGPSigned)
+  auto inputType = input.type ();
+
+  if (inputType == Data::Type::PGPSigned)
     {
       verify = true;
       decrypt = false;
@@ -281,7 +291,7 @@ ParseController::parse()
              decrypt, verify,
              protocol == OpenPGP ? "OpenPGP" :
              protocol == CMS ? "CMS" : "Unknown",
-             m_sender.empty() ? "none" : m_sender.c_str(), input.type ());
+             m_sender.empty() ? "none" : m_sender.c_str(), inputType);
   if (decrypt)
     {
       input.seek (0, SEEK_SET);
@@ -310,7 +320,7 @@ ParseController::parse()
         {
           verify = false;
         }
-      if (m_decrypt_result.error())
+      if (m_decrypt_result.error () || m_decrypt_result.isNull ())
         {
           m_error = format_error (m_decrypt_result, protocol);
         }
@@ -465,6 +475,7 @@ ParseController::get_ultimate_keys()
     }
   ctx->setKeyListMode (KeyListMode::Local);
   Error err;
+  TRACEPOINT;
   if ((err = ctx->startKeyListing ()))
     {
       log_error ("%s:%s: Failed to start keylisting err: %i: %s",
@@ -473,23 +484,36 @@ ParseController::get_ultimate_keys()
       gpgrt_lock_unlock (&keylist_lock);
       return s_ultimate_keys;
     }
+  TRACEPOINT;
   while (!err)
     {
       const auto key = ctx->nextKey(err);
-      if (key.isNull() || err)
+      if (err || key.isNull())
         {
+          TRACEPOINT;
           break;
         }
+      if (key.isInvalid ())
+        {
+          log_debug ("%s:%s: skipping invalid key.",
+                     SRCNAME, __func__);
+          continue;
+        }
       for (const auto uid: key.userIDs())
         {
-          if (uid.validity() == UserID::Validity::Ultimate)
+          if (uid.validity() == UserID::Validity::Ultimate &&
+              uid.id())
             {
               s_ultimate_keys.push_back (key);
-              log_debug ("adding ultimate uid: %s", uid.id());
+              log_debug ("%s:%s: Adding ultimate uid.",
+                         SRCNAME, __func__);
+              log_mime_parser ("%s:%s: Added uid %s.",
+                               SRCNAME, __func__, uid.id());
               break;
             }
         }
     }
+  TRACEPOINT;
   delete ctx;
   log_debug ("%s:%s: keylisting done.",
              SRCNAME, __func__);
