@@ -1333,7 +1333,16 @@ mark_mime_action (LPDISPATCH ctrl, int flags, bool is_explorer)
 
   oldflags = get_gpgol_draft_info_flags (message);
 
-  newflags = oldflags xor flags;
+  if (flags == 3 && oldflags != 3)
+    {
+      // If only one sub button is active activate
+      // both now.
+      newflags = 3;
+    }
+  else
+    {
+      newflags = oldflags xor flags;
+    }
 
   if (set_gpgol_draft_info_flags (message, newflags))
     {
@@ -1348,6 +1357,11 @@ mark_mime_action (LPDISPATCH ctrl, int flags, bool is_explorer)
       we invalidate a lot *sigh* */
   gpgoladdin_invalidate_ui ();
 
+  if (newflags & 1)
+    {
+      Mail::locate_all_crypto_recipients ();
+    }
+
 done:
   gpgol_release (context);
   gpgol_release (mailitem);
@@ -1458,29 +1472,7 @@ get_mail_from_control (LPDISPATCH ctrl, bool *none_selected)
     }
   else if (!strcmp (ctx_name.c_str(), "_Explorer"))
     {
-      LPDISPATCH selection = get_oom_object (context, "Selection");
-      if (!selection)
-        {
-          log_error ("%s:%s: Failed to get selection.",
-                     SRCNAME, __func__);
-          gpgol_release (context);
-          return NULL;
-        }
-      int count = get_oom_int (selection, "Count");
-      if (count == 1)
-        {
-          // If we call this on a selection with more items
-          // Outlook sends an ItemLoad event for each mail
-          // in that selection.
-          mailitem = get_oom_object (selection, "Item(1)");
-        }
-      gpgol_release (selection);
-
-      if (!mailitem)
-        {
-          *none_selected = true;
-        }
-      else if (g_ol_version_major >= 16)
+      if (g_ol_version_major >= 16)
         {
           // Avoid showing wrong crypto state if we don't have a reading
           // pane. In that case the parser will finish for a mail which is gone
@@ -1526,6 +1518,35 @@ get_mail_from_control (LPDISPATCH ctrl, bool *none_selected)
                 }
             }
         }
+      if (!*none_selected)
+        {
+          /* Accessing the selection item can trigger a load event
+             so we only do this here if we think that there might be
+             something visible / selected. To avoid triggering a load
+             if there is no content shown. */
+          LPDISPATCH selection = get_oom_object (context, "Selection");
+          if (!selection)
+            {
+              log_error ("%s:%s: Failed to get selection.",
+                         SRCNAME, __func__);
+              gpgol_release (context);
+              return NULL;
+            }
+          int count = get_oom_int (selection, "Count");
+          if (count == 1)
+            {
+              // If we call this on a selection with more items
+              // Outlook sends an ItemLoad event for each mail
+              // in that selection.
+              mailitem = get_oom_object (selection, "Item(1)");
+            }
+          gpgol_release (selection);
+
+          if (!mailitem)
+            {
+              *none_selected = true;
+            }
+        }
     }
 
   gpgol_release (context);
@@ -1584,6 +1605,12 @@ HRESULT get_is_details_enabled (LPDISPATCH ctrl, VARIANT *result)
 {
   MY_MAIL_GETTER
 
+  if (!result)
+    {
+      TRACEPOINT;
+      return S_OK;
+    }
+
   result->vt = VT_BOOL | VT_BYREF;
   result->pboolVal = (VARIANT_BOOL*) xmalloc (sizeof (VARIANT_BOOL));
   *(result->pboolVal) = none_selected ? VARIANT_FALSE : VARIANT_TRUE;
@@ -1713,15 +1740,20 @@ HRESULT launch_cert_details (LPDISPATCH ctrl)
               log_error ("%s:%s: No spawn engine.",
                          SRCNAME, __func__);
             }
+            std::string parentWid = std::to_string ((int) (intptr_t) get_active_hwnd ());
             const char *argv[] = {path.c_str(),
                                   "--query",
                                   mail->get_sig_fpr(),
+                                  "--parent-windowid",
+                                  parentWid.c_str(),
                                   NULL };
             log_debug ("%s:%s: Starting %s %s %s",
                        SRCNAME, __func__, path.c_str(), argv[1], argv[2]);
             Data d(Data::null);
             ctx->spawnAsync(path.c_str(), argv, d, d,
-                            d, Context::SpawnNone);
+                            d, (GpgME::Context::SpawnFlags) (
+                                GpgME::Context::SpawnAllowSetFg |
+                                GpgME::Context::SpawnShowWindow));
         }
       else
         {
