@@ -22,6 +22,10 @@
 
 #include <windows.h>
 
+#include "config.h"
+
+#include <gpg-error.h>
+
 /** Window Message handling for GpgOL.
   In Outlook only one thread has access to the Outlook Object model
   and this is the UI Thread. We can work in other threads but
@@ -36,13 +40,18 @@
   */
 typedef enum _gpgol_wmsg_type
 {
-  UNKNOWN = 0,
-  INVALIDATE_UI = 1, /* The UI should be invalidated. */
-  PARSING_DONE = 2, /* A mail was parsed. Data should be a pointer
+  UNKNOWN = 1100, /* A large offset to avoid conflicts */
+  INVALIDATE_UI, /* The UI should be invalidated. */
+  PARSING_DONE, /* A mail was parsed. Data should be a pointer
                       to the mail object. */
-  RECIPIENT_ADDED = 3, /* A recipient was added. Data should be ptr
-                          to mail */
-  CLOSE = 4, /* Send the message in the next event loop. */
+  RECIPIENT_ADDED, /* A recipient was added. Data should be ptr
+                      to mail */
+  CLOSE, /* Close the message in the next event loop. */
+  CRYPTO_DONE, /* Sign / Encrypt done. */
+  WKS_NOTIFY, /* Show a WKS Notification. */
+  BRING_TO_FRONT, /* Bring the active Outlook window to the front. */
+  INVALIDATE_LAST_MAIL,
+  REVERT_MAIL,
 } gpgol_wmsg_type;
 
 typedef struct
@@ -67,6 +76,11 @@ send_msg_to_ui_thread (wm_ctx_t *ctx);
 int
 do_in_ui_thread (gpgol_wmsg_type type, void *data);
 
+/** Send a message to the UI thread but returns
+    immediately without waiting for the execution. */
+void
+do_in_ui_thread_async (gpgol_wmsg_type type, void *data);
+
 /** Create our filter before outlook Window Messages. */
 HHOOK
 create_message_hook();
@@ -80,4 +94,7 @@ close_mail (LPVOID);
 void add_explorer (LPDISPATCH explorer);
 void remove_explorer (LPDISPATCH explorer);
 
+/* The lock to invalide the ui */
+extern gpgrt_lock_t invalidate_lock;
+
 #endif // WINDOWMESSAGES_H
